#!/bin/bash
#
# Adminstration script for the jumphosts
#
# The script is used to set up and administer users.
# Run with '-h' to get a short usage message.
#

# Exit codes
ENORMAL=0
EINVALOPT=1
EINVALCMD=2
ENOUSER=3
EUSEREXIST=4
EUSERLOGGEDON=5
EUNKNOWN=6

# check which RHEL version this is
version=$( cat /etc/redhat-release | grep -oP "[0-9]+" | head -1 )

# access file on RHEL 8 and greater
accessfile="/etc/security/access.d/66-jumpusers.conf"
default_accesslist="129.240.0.0/16 158.36.5.0/24 193.157.128.0/17 2001:700:100::/40"

usage()
{
    cat << EOF

$0 - Administrating jump-hosts

Usage:

  $0 -h|<command> <user> [...]

  Commands:

    add <user> [full name]  : add local user to systems (user does not exist)
    remove <user>           : remove user from systems
    new_ga <user>           : generate a (new) Google Authenticator configuration for an existing user (either local or LDAP)
    otp <user> [num]        : print "num" one-time password(s) for user, default num = 5
    gen_otp <user> [num]    : generate new emergency scratch codes for user, default num = 1
    secret <user>           : retrieve the secret for user, use this when configuring 2factor applications

EOF

}


user_dir()
{
    local dir
    local user=$1
    dir=$(getent passwd $user | cut -d: -f6)
    if [ -z $dir -o $dir == "/" ]; then
        local i
        i=$(grep HOME= /etc/default/useradd)
        dir=${i#*=}/$user
    fi
    echo $dir
 }

config_dir()
{
   echo $(user_dir $1)
#    echo /var/lib/google-authenticator
}

# Routine to generate new Google Authenticator user configuration
# (including secret and One Time passwords)
setup_ga()
{
   user=$1
   userdir=$(user_dir $user)


   if [ -z $user ]; then
       echo
       echo "username must be provided!"
       exit $ENOUSER
   fi

   # check that user already exists
   if [ $version -ge 8 ]; then
       grep $user $accessfile
   else
       id $user >/dev/null 2>&1
   fi
   if [ $? -ne 0 ]; then
       echo
       echo "$user has not been defined on this host!"
       exit $ENOUSER
   fi

   echo "Generating configuration for 2FA"
   su - $user -c "/usr/bin/google-authenticator -tdr3 -R30 -w3 -f"
   if [ ! -d $userdir/.ssh ]; then
      # should really check that this succeeds (.ssh migh be a file for instance
      su - $user -c "mkdir $userdir/.ssh"
   fi
   mv -f $userdir/.google_authenticator $userdir/.ssh
   # fix: if only 80 bit key generated extend to 160 bit
   ed << EOF $userdir/.ssh/.google_authenticator
1s/^\([a-zA-Z0-9]\{26\}\)$/\1AAAAAA/
w
q
EOF
   restorecon -R $userdir
}

# add new user to the systems
# creating account if non-existent and sets up the totp infrastructure
# username as first argument, full name as second argument (optional)
#
add()
{
    user=$1
    fullname=$2
    userdir=$(user_dir $user)

    if [ -z "$fullname" ]; then fullname=""; fi

    if [ $version -ge 8 ]; then
        # on RHEL 8 every user is knwon on the system - check if access is granted
        grep -qs $user $accessfile
        if [ $? ]; then
            echo
            echo "$user does not exist, adding!"
            echo "+ : $user : $default_accesslist" >> $accessfile
            mkdir -p $userdir
            chown $user:$user $userdir
        fi
    else
        # on rhel 7 only explicitly configured users are known - does user already exist
        id $user >/dev/null 2>&1
        if [ $? -ne 0 ]; then
            echo
            echo "$user does not exist, adding!"
            if [ $(getent group $user) ]; then
                useradd -c "$fullname" -m -g $user -G wheel $user
            else
                useradd -c "$fullname" -m -U -G wheel $user
            fi
            echo "Set password for $user when prompted."
            echo
            passwd $user
        fi
    fi

    su - $user -c "echo 'y
n
n
y
' | /usr/bin/google-authenticator -tdr3 -R30 -w3 -f"
    if [ ! -d "$userdir/.ssh" ]; then su - $user -c "mkdir $userdir/.ssh"; fi
#    [ ! -d $CONFIG ] || mkdir -p "$CONFIG"
    mv $userdir/.google_authenticator $userdir/.ssh
    restorecon -R $userdir
#    chown root:root $CONFIG

    # fix: if only 80 bit key generated extend to 160 bit
    ed << EOF $CONFIG >/dev/null 2>&1
1s/^\([a-zA-Z0-9]\{16\}\)$/\1AAAAAAAAAAAAAAAA/
w
q
EOF

    echo
    echo "=================================================="
    echo
    echo "User $user set up."
    echo "For the TOTP client application enter this secret:"
    echo
    head -1 $CONFIG
    echo
    echo "=================================================="
    echo
}

# removes a user from the systems, including the totp setup
# leaves the home directory intact (to save any potential essential files)
#
remove()
{
    ret=0
    user=$1

    echo "Removing $user from jump-hosts"
    echo
    if [ $version -ge 8 ]; then
        # rhel 8

        grep -qs $user $accessfile || ret=6     # user does not exist
        who | grep -q $user && ret=8            # user is still ogged on

        if [ $ret == 0 ]; then
            sed -i "/$user/d" $accessfile
        fi
    else
        # rhel 7
	    userdel $user >/dev/null 2>&1
        ret=$?
    fi

    # handle error cases
    case "$ret" in
        0)  ;;
        6)
            echo "$user does not exist on system!"
            exit $ENOUSER
            ;;

        8)
            echo "$user still logged on, logg him/her off first"
            exit $EUSERLOGGEDON
            ;;
        *)
            echo "Unknown error when removing user with \`userdel\`: $?"
            exit $EUNKNOWN
            ;;
    esac

    if [ -d $HOMEDIR ]; then
        echo "== Not removing home directory ($HOMEDIR) or mailspool (/var/spool/mail/$user)!"
        echo "== Check contents and remove manually"
        echo
    fi
    rm $CONFIG
}

# retrieves and print to standard out a number of one time passwords
# username as first argument
# number of passwords to get as second argument (5 is standard)
# NB: there might be fewer otp's available than requested
#
get_otp()
{
    user=$1
    no=$2

    if [ -z $no ]; then no=5; fi

    if [ -f $CONFIG ]; then
        echo "One Time passwords for $user ($no requested):"
        echo
        grep -E -m $no "^[0-9]{8}$" $CONFIG
    else
        echo "There does not seem to be any OTP setup for $user!"
        exit $ENOUSER
    fi
}

# generates new emergency scratch codes for given user
# username as first argument
# number of codes to generate as second argument (max 25)
#
generate_otp()
{
    user=$1
    no=$2

    if [ -z $no ]; then no=1; fi
    if [ $no -gt 15 ]; then no=15; fi

    if [ ! -f "$CONFIG" ]; then
        echo "$user does not seem to be set up properly on this system!"
        exit $ENOUSER
    fi

    echo "Generating new emergency scratch codes for $user..."
    if [ $no -gt 3 ]; then
        echo "This will take approx. $no seconds"
    fi
    echo
    secret=`head -1 $CONFIG`
    for i in `seq $no`; do
        echo -n "."
        oathtool --totp -b $secret -d8 -s1 >> $CONFIG
        sleep 1
    done
    echo
    echo "Done"
}

# get the secret for given user (needed to set up a 2 factor authentication application)
# username as first argument
#
get_secret()
{
    user=$1

    if [ ! -f "$CONFIG" ]; then
        echo "$user does not seem to be set up properly on this system!"
        exit $ENOUSER
    fi
    secret=`head -1 $CONFIG`
    echo
    echo "Secret for $user: $secret"
    echo "Use this when setting up the TOTP application"
    echo
}


#
# Main part
#

while getopts ":h" opt; do
    case $opt in

        h)
            usage
            exit $ENORMAL
            ;;

        \?)
            echo "Unknown parameter!"
            usage
            exit $EINVALOPT
            ;;

    esac
done

shift $((OPTIND-1))
command=$1;
shift
user=$1
HOMEDIR=$(config_dir $user)
CONFIG=$HOMEDIR/.ssh/.google_authenticator
#CONFIG=$HOMEDIR/$user

case $command in

    "add")
        shift
        fullname=$1
        add $user "$fullname"
        exit $ENORMAL
        ;;

    "remove")
        remove $user
        exit $ENORMAL
        ;;

    "new_ga")
        setup_ga $user
        get_secret $user
        exit $ENORMAL
        ;;

    "otp")
        shift
        no=$1
        get_otp $user $no
        exit $ENORMAL
        ;;

    "gen_otp")
        shift
        no=$1
        generate_otp $user $no
        exit $ENORMAL
        ;;

    "secret")
        shift
        no=$1
        get_secret $user
        exit $ENORMAL
        ;;

    *)
        echo "Unknown command!"
        usage
        exit $EINVALCMD
        ;;
esac

